#if 0
# This file is generated by cbindgen. DO NOT EDIT
#endif


#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

void root(void);

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#if 0
# This is a simple test to ensure that trailers do not cause extra newlines in files
#endif
