#include <cstdarg>
#include <cstdint>
#include <cstdlib>
#include <ostream>
#include <new>

static const uint8_t A = 0;

static const uint8_t B = 0;

extern "C" {

extern const uint8_t C;

extern const uint8_t D;

} // extern "C"
