from libc.stdint cimport int8_t, int16_t, int32_t, int64_t, intptr_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, uintptr_t
cdef extern from *:
  ctypedef bint bool
  ctypedef struct va_list

cdef extern from *:

  void ptr_as_array(uint32_t n, const uint32_t arg[3], const uint64_t *v);

  void ptr_as_array1(uint32_t n, const uint32_t arg[3], uint64_t v[4]);

  void ptr_as_array2(uint32_t n, uint32_t arg[], const uint64_t v[]);

  void ptr_as_array_wrong_syntax(uint32_t *arg, const uint32_t *v, const uint32_t*);

  void ptr_as_array_unnamed(uint32_t*, const uint32_t*);
